from item_model import ItemModel
from typing import Union
import numpy as np
from logistic_2PM import Logistic2PM

class Logistic1PModel(ItemModel):
    
    def prob(self,
             ability: Union[list, np.ndarray, float], 
             params: Union[list, np.ndarray])  -> np.ndarray:
        return Logistic2PM().prob(ability, [params[0], np.ones(len(params[0]))])

    def estimate_parameters(self, 
                            ability: Union[list, np.ndarray], 
                            result: Union[list, np.ndarray], 
                            sigm_orig=0,
                            end=0.0000001,  
                            eps=0.0001) -> np.ndarray:
        ability = np.array(ability)
        result = np.array(result)
        est = sigm_orig
        prev_est = est + 2 * end
        while abs(est - prev_est) > end:
            P = self.prob(
                ability, [[est]]).transpose()[0]
            L1 = np.sum(result - P)
            L2 = -np.sum(P * (1 - P))

            if abs(L2) < eps:
                break
            
            prev_est = est
            est = est - L1 / L2
        return np.array([est])

    def estimate_ability(self,
                         params: Union[list, np.ndarray],
                         results: Union[list, np.ndarray],
                         end=0.00000001,
                         eps=0.01) -> np.ndarray:
        return Logistic2PM().estimate_ability([params[0], np.ones(len(params[0]))], results, end, eps)




from item_model import ItemModel
from typing import Union
import numpy as np
import matplotlib.pyplot as plt

class PL2(ItemModel):
    def prob(self, 
             ability: Union[list, np.ndarray, float], 
             params: Union[list, np.ndarray])  -> np.ndarray:
        items = np.array(params).shape[1]
        num_abilities = 1
        if np.array(ability).shape != ():
            num_abilities = np.array(ability).shape[0]
        ability = np.tile(np.array([ability]).transpose(), (1, items))

        params = np.tile(params, num_abilities).reshape(
            (2, num_abilities, items))
        return 1 / (1 + np.exp(-params[0] - (params[1] * ability)))

    def estimate_parameters(self, 
                            ability: Union[list, np.ndarray], 
                            result: Union[list, np.ndarray], 
                            sigm_orig=-1, 
                            lamb_orig=1, 
                            end=0.00001,  
                            eps=0.1) -> np.ndarray:
        ability = np.array(ability)
        result = np.array(result)
        est = [sigm_orig, lamb_orig]
        prev_est = [0, 0]
        while abs(est[0] - prev_est[0]) > end or abs(est[1] - prev_est[1]) > end:
            P = self.prob(
                ability, [[est[0]], [est[1]]]).transpose()[0]
            W = P * (1 - P)
            L11 = -np.sum(W)
            L12 = -np.sum(ability * W)
            L22 = -np.sum(ability**2 * W)
            L = np.array([[L11, L12], [L12, L22]])
            if abs(np.linalg.det(L)) < eps:
                break
            L_inv = np.linalg.inv(L)

            obs_mat = np.array(
                [np.sum(result - P), np.sum((result - P) * ability)])
            prev_est = est
            est = est - np.matmul(L_inv, obs_mat)
        return est

    def estimate_ability(self,
                         params: Union[list, np.ndarray],
                         results: Union[list, np.ndarray],
                         end=0.00000001,
                         eps=0.01) -> np.ndarray:
        params = np.array(params)
        results = np.array(results)
        est = 0.5
        prev_est = 0
        while abs(est - prev_est) > end:
            P = self.prob(est, params)[0]
            W = (1 - P) * P
            denom = np.sum(params[1]**2 * W)
            if abs(denom) < eps or np.any(np.abs(W) < eps):
                break            
            prev_est = est
            est = est + (np.sum(params[1] * W  * ((results - P)/ W)) / denom)
        return est
    
    def convert_param_form(self, params):
        return np.array([-params[0] / params[1], 1 / params[1]])
    
    
    # def train_em_mle(self, data) -> tuple:
    #     return super().train_em_mle(data)
    
    # def plot_item_curve(self, p):
    #     ability = np.linspace(-4, 4, 100)
    #     for i in range(p.shape[1]):
    #         plt.plot(ability, self.prob(ability, p[:,i].reshape(p.shape[0], 1)), label=f'Question{i+1}')        
    #     plt.legend()
    #     plt.savefig('item_curve.png')
        


from item_model import ItemModel
from typing import Union
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize    


class Logistic3PM(ItemModel):
    
    def prob(self, 
             ability: Union[list, np.ndarray, float], 
             params: Union[list, np.ndarray])  -> np.ndarray:
        items = np.array(params).shape[1]
        num_abilities = 1
        if np.array(ability).shape != ():
            num_abilities = np.array(ability).shape[0]
        ability = np.tile(np.array([ability]).transpose(), (1, items))

        params = np.tile(params, num_abilities).reshape(
            (3, num_abilities, items))
        return params[2] + (1 - params[2]) / (1 + np.exp(-params[0] * (ability - params[1])))
    
    def prob2PM(self,
                ability: Union[list, np.ndarray, float],
                params: Union[list, np.ndarray]) -> np.ndarray:
        items = np.array(params).shape[1]
        num_abilities = 1
        if np.array(ability).shape != ():
            num_abilities = np.array(ability).shape[0]
        ability = np.tile(np.array([ability]).transpose(), (1, items))

        params = np.tile(params, num_abilities).reshape(
            (3, num_abilities, items))
        return 1 / (1 + np.exp(-params[0] * (ability - params[1])))
    
    def estimate_parameters(self, 
                            ability: Union[list, np.ndarray], 
                            result: Union[list, np.ndarray],
                            a_orig=1,
                            b_orig=0.1, 
                            c_orig=0.1,
                            end=0.0000001,  
                            eps=0.1) -> np.ndarray:
        # use orig_params?
        ability = np.array(ability)
        result = np.array(result)
        est = np.array([a_orig, b_orig, c_orig])
        prev_est = est + 2 * end
        while np.all(np.abs(prev_est - est) > end):
            # self.plot_item_curve(est.reshape((3,1)))
            P = self.prob(ability, est.reshape((3,1))).T[0]
            Q = 1 - P
            P_2pm = self.prob2PM(ability, est.reshape((3,1))).T[0]
            
            L11 = -np.sum((ability-est[1])**2 * P * Q * (P_2pm/P)**2)
            L12 = np.sum(est[0] * (ability - est[1]) * P * Q * (P_2pm/P))
            L13 = -np.sum((ability - est[1]) * (Q / (1 - est[2])) * (P_2pm/P))
            
            L22 = -est[0]**2 * np.sum(P * Q * (P_2pm/P))
            L23 = np.sum(est[0] * (Q / (1 - est[2])) * (P_2pm/P))
            
            L33 = -np.sum((Q / (1 - est[2])) / (P - est[2]) * (P_2pm/P))

            L = np.array([[L11, L12, L13], [L12, L22, L23], [L13, L23, L33]])
            
            if abs(np.linalg.det(L)) < eps:
                print('break')
                break
            L_inv = np.linalg.inv(L)
            
            L1 = np.sum((result - P) * (ability - est[1]) * (P_2pm/P))
            L2 = -est[0] * np.sum((result - P) * (P_2pm/P))
            L3 = np.sum((result - P) / (P - est[2]) * (P_2pm/P))
            obs_mat = np.array([L1, L2, L3])
            prev_est = est
            est = est - np.matmul(L_inv, obs_mat)
        return est
    
    def estimate_ability(self,
                         params: Union[list, np.ndarray],
                         results: Union[list, np.ndarray],
                         end=0.00000001,
                         eps=0.01) -> np.ndarray:
        params = np.array(params)
        results = np.array(results)
        est = 0.5
        prev_est = 0
        while abs(est - prev_est) > end:
            P = self.prob(est, params)[0]
            Q = 1 - P
            P_2pm = self.prob2PM(est, params)[0]
            W = P_2pm * (1 - P_2pm)
            
            denom = - np.sum(params[0]**2 * W * (P_2pm/P)**2)
            
            print(denom)
            if abs(denom) < eps:
                break
            num = np.sum(params[0] * W * ((results - P) / (P * Q)) * (P_2pm/P))
            prev_est = est
            est = est + (num / denom)
        return est
    
    def plot_item_curve(self, p):
        ability = np.linspace(-5, 5, 100)
        for i in range(p.shape[1]):
            plt.plot(ability, self.prob(ability, p[:,i].reshape(p.shape[0], 1)), label=f'Question{i+1}')        
        plt.legend()
        plt.show()
